/* --------------------------------------------------------------------------
 * module_alsamixer.c
 * code for the ALSA mixer module, controling volume levels
 *
 * Copyright 2004 John Steele Scott <toojays@toojays.net>
 * Copyright 2002 Matthias Grimm
 * Copyright 1999-2000 by Jaroslav Kysela <perex@suse.cz>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 * -------------------------------------------------------------------------*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef WITH_ALSA

#include <alsa/asoundlib.h>
#include "pbbinput.h"

#include <math.h>
#include <pbb.h>

#include "gettext_macros.h"
#include "input_manager.h"
#include "module_alsamixer.h"
#include "module_ossmixer.h"
#include "support.h"

struct moddata_alsamixer {
	char *card;	 /* name of the soundcard (usually "default" is fine) */
	char *channels;		/* the "MixerChannels" config string */
	unsigned short keyvolup;
	unsigned short modvolup;
	unsigned short keyvoldn;
	unsigned short modvoldn;
	unsigned short keymute;
	unsigned short modmute;
	int init_complete;
	int mixer_delayed;
	int mute;
	snd_mixer_t * mixer;
	snd_mixer_elem_t ** elements; /* points to a NULL terminated list, first is master */
} modbase_alsamixer;

/* initialise the list of elements, based on the "channels" string */
static int
alsamixer_setup_elements_internal (char * channels, snd_mixer_selem_id_t *sid)
{
	struct moddata_alsamixer *base = &modbase_alsamixer;
    char *token;
	char buffer[STDBUFFERLEN];
	int n;

	/* get rid of old list */
	free (base->elements);
	base->elements = NULL;
	base->init_complete = 0;
	n = 0;

	/* this copy we save for returning in alsamixer_handle_tags */
	strncpy (base->channels, channels, STDBUFFERLEN);

	/* this copy we use for strtok */
	strncpy (buffer, channels, STDBUFFERLEN);
	cleanup_buffer (buffer);
	if ((token = strtok(buffer, ",\n"))) {
		do {
			/* ensure we have enough memory to store pointer + terminator */
			base->elements = realloc (base->elements, (n+2)*sizeof (snd_mixer_elem_t *));
			if (base->elements == NULL) {
				print_error (_("ERROR: memory allocation failed\n"));
				return -1;
			}

			/* find element */
			for (base->elements[n] = snd_mixer_first_elem (base->mixer);
			     base->elements[n] != NULL;
			     base->elements[n] = snd_mixer_elem_next (base->elements[n])) {
				snd_mixer_selem_get_id (base->elements[n], sid);
				if (strcasecmp (snd_mixer_selem_id_get_name (sid), token) == 0)
					break;
			}
			
			if (base->elements[n] == NULL) {
				/* report any element we can't find */
				print_error (_("WARNING: card \"%s\" has no \"%s\" element\n"), base->card, token);
			} else {
				n++;
			}
			
		} while ((token = strtok(0,",\n")) != NULL);

		/* terminate list */
		base->elements[n] = NULL;

		/* check that we have a master element */
		if (!base->elements[0]) {
			print_error (_("WARNING: MixerChannels contains no valid elements\n"));
			return -1;
		}

		/* check that master element has a playback volume */
		if (!snd_mixer_selem_has_playback_volume (base->elements[0])) {
			snd_mixer_selem_get_id (base->elements[n], sid);
			print_error (_("WARNING: mixer element \"%s\" has no playback volume control\n"), snd_mixer_selem_id_get_name (sid));
			return -1;
		}

		base->init_complete = 1;
		return 0;
	} else {
		/* no tokens? */
		return -1;
	}
}

/* wrapper arround alsamixer_setup_elements_internal so we can be sure to free the sid. */
static int
alsamixer_setup_elements (char * channels)
{
	snd_mixer_selem_id_t *sid;
	int err, result;

	/* create a selem_id to extract names from */
	if ((err = snd_mixer_selem_id_malloc (&sid))){
		print_error (_("ERROR: failed memory allocation: %s\n"), snd_strerror (err));
		return -1;
	}

	result = alsamixer_setup_elements_internal (channels, sid);
	snd_mixer_selem_id_free (sid);
	return result;
}

static int
alsamixer_finish_init ()
{
	struct moddata_alsamixer *base = &modbase_alsamixer;
	int err;

	/* setup mixer */
	if ((err = snd_mixer_open (&base->mixer, 0)) < 0) {
		print_error (_("ERROR: mixer open %s: %s\n"), base->card, snd_strerror (err));
		return -1;
	}
	if ((err = snd_mixer_attach (base->mixer, base->card)) < 0) {
		print_error (_("ERROR: mixer attach %s: %s\n"), base->card, snd_strerror (err));
		snd_mixer_free (base->mixer);
		return -1;
	}
	if ((err = snd_mixer_selem_register (base->mixer, NULL, NULL)) < 0) {
		print_error (_("ERROR: mixer register: %s\n"), snd_strerror (err));
		snd_mixer_free (base->mixer);
		return -1;
	}
	
	if ((err = snd_mixer_load (base->mixer)) < 0) {
		print_error (_("ERROR: load %s: %s\n"), base->card, snd_strerror (err));
		snd_mixer_free (base->mixer);
		return -1;
	}

	return alsamixer_setup_elements (base->channels);
}

/* Converts an ALSA volume to a number in the range [0, +/-100]. This function
 * assumes that the mixer has already been setup successfully.
 */
static inline int
alsamixer_vol_to_percentage (snd_mixer_elem_t * elem, long volume)
{
	long range, vol_min, vol_max;
	int pct;

	snd_mixer_selem_get_playback_volume_range (elem, &vol_min, &vol_max);
	range = vol_max - vol_min;
	pct = rint(volume*100.0/range);
	return pct;
}

/* Converts a number in the range [0, +/-100] to an ALSA volume. This function
 * assumes that the mixer has already been setup successfully.
 */
static long
alsamixer_percentage_to_vol (snd_mixer_elem_t * elem, int volume)
{
	long range, vol_min, vol_max;
	
	snd_mixer_selem_get_playback_volume_range (elem, &vol_min, &vol_max);
	range = vol_max - vol_min;
	return rint(volume*range/100.0);
}

/* Saves the current mute status in base->mute, and returns the volume of the first playback channel of the master element */
static int
alsamixer_get_volume ()
{
	long volume;
	struct moddata_alsamixer *base = &modbase_alsamixer;
	snd_mixer_elem_t * elem;

	if (base->init_complete == 0) /* is mixer setup already completed? */
		if ((alsamixer_finish_init ()) != 0) /* no, then do it now */
			return -1; /* Oops, try it again later */

	/* force simple mixer to update its values from audio driver*/
	snd_mixer_handle_events (base->mixer);

	/* read values */
	elem = base->elements[0];
	snd_mixer_selem_get_playback_switch (elem, 0, &base->mute);
	base->mute = !base->mute;
	snd_mixer_selem_get_playback_volume (elem, 0, &volume);
	return volume;
}

/* same as alsamixer_get_volume, but returns the volume as number between 0 and 100 */
static int 
alsamixer_get_volume_as_percentage ()
{
	int volume;
	struct moddata_alsamixer *base = &modbase_alsamixer;

	volume = alsamixer_get_volume ();
	
	if (volume < 0)
		return volume;
	else
		return alsamixer_vol_to_percentage (base->elements[0], volume);
}

static void
alsamixer_set_and_send (int increment)
{
	long vol, vol_min, vol_max;
	long master_percentage, master_vol, master_vol_max, master_vol_min;
    snd_mixer_elem_t ** elem_ptr;
	snd_mixer_elem_t * elem;
	struct moddata_alsamixer *base = &modbase_alsamixer;

	if (base->init_complete == 0) /* is mixer setup already completed? */
		if ((alsamixer_finish_init ()) != 0) /* no, then do it now */
			return;	/* Oops, try it again later */


    /* get initial values from master element */
	elem_ptr = base->elements;
	elem = *elem_ptr;
	snd_mixer_selem_get_playback_volume_range (elem, &master_vol_min, &master_vol_max);
	master_vol = alsamixer_get_volume ();
	master_vol = master_vol + increment;
	if (master_vol > master_vol_max)
		master_vol = master_vol_max;
	else if (master_vol < master_vol_min)
		master_vol = master_vol_min;
	master_percentage = alsamixer_vol_to_percentage (elem, master_vol);			
	
	if (increment == 0)
		/* toggle mute status */
		base->mute = !base->mute;
	else
		/* volume change releases the mute */
		base->mute = 0;

	/* cycle through elements */
	for (elem_ptr = base->elements; *elem_ptr != NULL; elem_ptr++) {
		elem = *elem_ptr;

		/* sync playback switch */
		snd_mixer_selem_set_playback_switch_all (elem, !base->mute);
		
		/* sync volume */
		if (snd_mixer_selem_has_playback_volume (elem)) {
			
			/* compare volume limits with master */
			snd_mixer_selem_get_playback_volume_range (elem, &vol_min, &vol_max);
			if ((vol_min != master_vol_min) || (vol_max != master_vol_max))
				vol = alsamixer_percentage_to_vol (elem, master_percentage);
			else
				vol = master_vol;
		
			/* sync volume with master element */
			snd_mixer_selem_set_playback_volume_all (elem, vol);
                }
        }
                
	if (base->mute)
		singletag_to_clients (CHANGEVALUE, TAG_MUTE, 0);
	else
		singletag_to_clients (CHANGEVALUE, TAG_VOLUME, master_percentage);
}

static int
alsamixer_handle_tags (int cfgure, struct tagitem *taglist)
{
	struct moddata_alsamixer *base = &modbase_alsamixer;
	int volume_delta, rc;
	rc = 0;

	while (taglist->tag != TAG_END) {
		switch (taglist->tag) {
		case TAG_VOLUME:
			if (cfgure) {
				volume_delta = taglist->data - alsamixer_get_volume_as_percentage ();
				if (volume_delta != 0)
					alsamixer_set_and_send (volume_delta);
			} else
				taglist->data = alsamixer_get_volume_as_percentage ();
			break;
		case TAG_MUTE:
			if (cfgure) {
				if (base->mute != taglist->data)
					alsamixer_set_and_send (0);
			} else {
				/* update mute status */
				alsamixer_get_volume ();
				taglist->data = base->mute;
			}
			break;
		case TAG_MIXERCARD:
			if (cfgure) {
				char * cardname = (char *) taglist->data;
				if (strlen (cardname) < STDBUFFERLEN) {
					strncpy (base->card, cardname, STDBUFFERLEN);
					alsamixer_exit ();
					alsamixer_finish_init ();
				} else
					rc = tagerror (taglist, E_BUFOVL);
			} else
				taglist->data = (long) base->card;
			break;
		case TAG_VOLUMEUPKEY:
			if (cfgure)	base->keyvolup = taglist->data;
			else		taglist->data = base->keyvolup;
			break;
		case TAG_VOLUMEUPMOD:
			if (cfgure)	base->modvolup = taglist->data;
			else		taglist->data = base->modvolup;
			break;
		case TAG_VOLUMEDOWNKEY:
			if (cfgure)	base->keyvoldn = taglist->data;
			else		taglist->data = base->keyvoldn;
			break;
		case TAG_VOLUMEDOWNMOD:
			if (cfgure)	base->modvoldn = taglist->data;
			else		taglist->data = base->modvoldn;
			break;
		case TAG_MUTEKEY:
			if (cfgure)	base->keymute = taglist->data;
			else		taglist->data = base->keymute;
			break;
		case TAG_MUTEMOD:
			if (cfgure)	base->modmute = taglist->data;
			else		taglist->data = base->modmute;
			break;
		case TAG_MIXERINITDELAY:
			if (cfgure)	base->mixer_delayed  = taglist->data;
			else		taglist->data = base->mixer_delayed;
			break;
		case TAG_MIXERCHANNELS:
			if (cfgure)	alsamixer_setup_elements ((char *) taglist->data);
			else		taglist->data = (long) base->channels;
			break;
		case TAG_SOUNDSYSTEM:
			if (cfgure)	tagerror (taglist, E_NOWRITE);
			else 		taglist->data = SOUNDSYSTEM_ALSA;
			break;
		}
		taglist++;
	}
	return rc;
}

int
alsamixer_init (struct tagitem *taglist)
{
	struct moddata_alsamixer *base = &modbase_alsamixer;
	static char devbuffer_mixer[STDBUFFERLEN];
	static char mixer_channels[STDBUFFERLEN];
	char *cardname;
	char *channels;
	int rc;

	base->card		= devbuffer_mixer;
	base->channels		= mixer_channels;
	base->elements          = NULL;
	base->keyvolup		= KEY_VOLUMEUP;
	base->modvolup		= MOD_NONE;
	base->keyvoldn		= KEY_VOLUMEDOWN;
	base->modvoldn		= MOD_NONE;
	base->keymute		= KEY_MUTE;
	base->modmute		= MOD_NONE;
	base->mixer_delayed = 0;
	base->init_complete = 0;

#ifdef WITH_OSS
	if ((tagfind (taglist, TAG_SOUNDSYSTEM, SOUNDSYSTEM_OSS)) != SOUNDSYSTEM_ALSA)
		return 0;  /* sorry, OSS was choosen */
#endif
	
	cardname = (char *) tagfind (taglist, TAG_MIXERCARD, (long) "default");
	if (strlen (cardname) < STDBUFFERLEN)
		strncpy (base->card, cardname, STDBUFFERLEN);
	else
		return  E_BUFOVL;

	channels = (char *) tagfind (taglist, TAG_MIXERCHANNELS, (long) "Master");
	if (strlen (channels) < STDBUFFERLEN)
		strncpy (base->channels, channels, STDBUFFERLEN);
	else
		return  E_BUFOVL;

	if ((base->mixer_delayed = tagfind (taglist, TAG_MIXERINITDELAY, 0)) == 0) {
		alsamixer_finish_init ();
	} else {
		/* can't set volume with only partly initialized mixer */
		tagskip (taglist, TAG_VOLUME);
		tagskip (taglist, TAG_MUTE);
	}

	if ((rc = alsamixer_handle_tags (MODE_CONFIG, taglist)))
		return rc;

	register_function (KBDQUEUE, alsamixer_keyboard);
	register_function (QUERYQUEUE, alsamixer_query);
	register_function (CONFIGQUEUE, alsamixer_configure);
	return 0;
}

int
alsamixer_exit ()
{
	struct moddata_alsamixer *base = &modbase_alsamixer;
	free (base->elements);
	base->elements = NULL;
	snd_mixer_free (base->mixer);
	base->init_complete = 0;
	return 0;
}

void
alsamixer_keyboard (struct tagitem *taglist)
{
	struct moddata_alsamixer *base = &modbase_alsamixer;
	int code, value, mod, step;

	code = (int) tagfind (taglist, TAG_KEYCODE, 0);
	value = (int) tagfind (taglist, TAG_KEYREPEAT, 0);
	mod = (int) tagfind (taglist, TAG_MODIFIER, 0);

	if (value) {
		if ((code == base->keyvolup) && ((mod & ~MOD_SHIFT) == base->modvolup)) {
			step = (mod & MOD_SHIFT) ? +10 : +1;
		} else if ((code == base->keyvoldn) && ((mod & ~MOD_SHIFT) == base->modvoldn)) {
			step = (mod & MOD_SHIFT) ? -10 : -1;
		} else if ((code == base->keymute) && (mod == base->modmute) && (value == 1)) {
			step = 0;
		} else return;

		alsamixer_set_and_send(step);
	}
}

void
alsamixer_query (struct tagitem *taglist)
{
	alsamixer_handle_tags (MODE_QUERY, taglist);
}

void
alsamixer_configure (struct tagitem *taglist)
{
	alsamixer_handle_tags (MODE_CONFIG, taglist);
}

#endif /* WITH_ALSA */
